@@ -40,6 +40,12 @@
 
 #define DEFAULT_DELAY 12
 
+/**
+ * The number of tries to check for a wait condition before aborting.
+ * TODO(sissel): Make this tunable at runtime?
+ */
+#define MAX_TRIES 500
+
 static void _xdo_populate_charcode_map(xdo_t *xdo);
 static int _xdo_has_xtest(const xdo_t *xdo);
 
@@ -68,7 +74,11 @@ static int _is_success(const char *funcname, int code);
 static char _keysym_to_char(const char *keysym);
 
 /* Default to -1, initialize it when we need it */
-static Atom _NET_WM_PID = -1;
+static Atom atom_NET_WM_PID = -1;
+static Atom atom_NET_WM_NAME = -1;
+static Atom atom_WM_NAME = -1;
+static Atom atom_STRING = -1;
+static Atom atom_UTF8_STRING = -1;
 
 xdo_t* xdo_new(char *display_name) {
   Display *xdpy;
@@ -131,11 +141,13 @@ const char *xdo_version(void) {
 }
 
 int xdo_window_wait_for_map_state(const xdo_t *xdo, Window wid, int map_state) {
+  int tries = MAX_TRIES;
   XWindowAttributes attr;
   attr.map_state = IsUnmapped;
-  while (attr.map_state != map_state) {
+  while (tries > 0 && attr.map_state != map_state) {
     XGetWindowAttributes(xdo->xdpy, wid, &attr);
     usleep(30000); /* TODO(sissel): Use exponential backoff up to 1 second */
+    tries--;
   }
   return 0;
 }
@@ -154,6 +166,13 @@ int xdo_window_unmap(const xdo_t *xdo, Window wid) {
   return _is_success("XUnmapWindow", ret == 0);
 }
 
+int xdo_window_reparent(const xdo_t *xdo, Window wid_source, Window wid_target) {
+  int ret = 0;
+  ret = XReparentWindow(xdo->xdpy, wid_source, wid_target, 0, 0);
+  XFlush(xdo->xdpy);
+  return _is_success("XReparentWindow", ret == 0);
+}
+
 int xdo_get_window_location(const xdo_t *xdo, Window wid,
                             int *x_ret, int *y_ret, Screen **screen_ret) {
   int ret;
@@ -229,8 +248,13 @@ int xdo_window_translate_with_sizehint(const xdo_t *xdo, Window window,
     height += hints.base_height;
   }
 
-  *width_ret = width;
-  *height_ret = height;
+  if (width_ret != NULL) {
+    *width_ret = width;
+  }
+
+  if (height_ret != NULL) {
+    *height_ret = height;
+  }
 
   return XDO_SUCCESS;
 }
@@ -241,11 +265,20 @@ int xdo_window_setsize(const xdo_t *xdo, Window window, int width, int height, i
   int cw_flags = 0;
 
   if (flags & SIZE_USEHINTS) {
-    xdo_window_translate_with_sizehint(xdo, window, width, height,
-                                       &wc.width, &wc.height);
-  } else {
-    wc.width = width;
-    wc.height = height;
+    flags |= SIZE_USEHINTS_X | SIZE_USEHINTS_Y;
+  }
+
+  wc.width = width;
+  wc.height = height;
+
+  if (flags & SIZE_USEHINTS_X) {
+    xdo_window_translate_with_sizehint(xdo, window, width, height, &wc.width,
+                                       NULL);
+  }
+
+  if (flags & SIZE_USEHINTS_Y) {
+    xdo_window_translate_with_sizehint(xdo, window, width, height, NULL,
+                                       &wc.height);
   }
 
   if (width > 0) {
@@ -261,7 +294,19 @@ int xdo_window_setsize(const xdo_t *xdo, Window window, int width, int height, i
   return _is_success("XConfigureWindow", ret == 0);
 }
 
-int xdo_window_setclass (const xdo_t *xdo, Window wid, const char *name, const char *class) {
+int xdo_window_set_override_redirect(const xdo_t *xdo, Window wid,
+                                     int override_redirect) {
+  int ret;
+  XSetWindowAttributes wattr;
+  long mask = CWOverrideRedirect;
+  wattr.override_redirect = override_redirect;
+  ret = XChangeWindowAttributes(xdo->xdpy, wid, mask, &wattr);
+
+  return _is_success("XChangeWindowAttributes", ret == 0);
+}
+
+int xdo_window_setclass (const xdo_t *xdo, Window wid, const char *name,
+                         const char *class) {
   int ret = 0;
   XClassHint *hint = XAllocClassHint();
   XGetClassHint(xdo->xdpy, wid, hint);
@@ -306,23 +351,41 @@ int xdo_window_focus(const xdo_t *xdo, Window wid) {
   return _is_success("XSetInputFocus", ret == 0);
 }
 
-int xdo_window_wait_for_size(const xdo_t *xdo, Window window, unsigned int width,
-                             unsigned int height, int flags, int to_or_from) {
-  unsigned int cur_width, cur_height;
+int xdo_window_wait_for_size(const xdo_t *xdo, Window window,
+                             unsigned int width, unsigned int height,
+                             int flags, int to_or_from) {
+  unsigned int cur_width = 0, cur_height = 0;
+  unsigned int alt_width = 0, alt_height = 0;
 
+  //printf("Want: %udx%ud\n", width, height);
   if (flags & SIZE_USEHINTS) {
     xdo_window_translate_with_sizehint(xdo, window, width, height,
-                                       &width, &height);
+                                       (int *)&width, (int *)&height);
+  } else {
+    unsigned int hint_width, hint_height;
+    /* TODO(sissel): fix compiler warning here, but it will require
+     * an ABI breakage by changing types... */
+    xdo_window_translate_with_sizehint(xdo, window, 1, 1,
+                                       &hint_width, &hint_height);
+    //printf("Hint: %dx%d\n", hint_width, hint_height);
+    /* Find the nearest multiple (rounded down) of the hint height. */
+    alt_width = (width - (width % hint_width));
+    alt_height = (height - (height % hint_height));
+    //printf("Alt: %udx%ud\n", alt_width, alt_height);
   }
 
+  int tries = MAX_TRIES;
   xdo_get_window_size(xdo, window, (unsigned int *)&cur_width,
                       (unsigned int *)&cur_height);
-  while (to_or_from == SIZE_TO
+  //printf("Want: %udx%ud\n", width, height);
+  //printf("Alt: %udx%ud\n", alt_width, alt_height);
+  while (tries > 0 && (to_or_from == SIZE_TO
          ? (cur_width != width && cur_height != height)
-         : (cur_width == width && cur_height == height)) {
+         : (cur_width == width && cur_height == height))) {
     xdo_get_window_size(xdo, window, (unsigned int *)&cur_width,
                         (unsigned int *)&cur_height);
     usleep(30000);
+    tries--;
   }
 
   return 0;
@@ -331,15 +394,18 @@ int xdo_window_wait_for_size(const xdo_t *xdo, Window window, unsigned int width
 int xdo_window_wait_for_active(const xdo_t *xdo, Window window, int active) {
   Window activewin = 0;
   int ret = 0;
+  int tries = MAX_TRIES;
 
   /* If active is true, wait until activewin is our window
    * otherwise, wait until activewin is not our window */
-  while (active ? activewin != window : activewin == window) {
+  while (tries > 0 && 
+         (active ? activewin != window : activewin == window)) {
     ret = xdo_window_get_active(xdo, &activewin);
     if (ret == XDO_ERROR) {
       return ret;
     }
     usleep(30000);
+    tries--;
   }
 
   return 0;
@@ -602,6 +668,44 @@ int xdo_window_get_active(const xdo_t *xdo, Window *window_ret) {
                      *window_ret == 0);
 }
 
+int xdo_window_select_with_click(const xdo_t *xdo, Window *window_ret) {
+  int screen_num;
+  Screen *screen;
+  xdo_mouselocation(xdo, NULL, NULL, &screen_num);
+
+  screen = ScreenOfDisplay(xdo->xdpy, screen_num);
+
+  /* Grab sync mode so we can ensure nothing changes while we figure
+   * out what the client window is.
+   * Also, everyone else who does 'select window' does it this way.
+   */
+  int grab_ret = 0;
+  grab_ret = XGrabPointer(xdo->xdpy, screen->root, False, ButtonReleaseMask,
+               GrabModeSync, GrabModeAsync, screen->root, None, CurrentTime);
+  if (grab_ret == AlreadyGrabbed) {
+    fprintf(stderr, "Attempt to grab the mouse failed. Something already has"
+            " the mouse grabbed. This can happen if you are dragging something"
+            " or if there is a popup currently shown\n");
+    return XDO_ERROR;
+  }
+  
+  XEvent e;
+  XAllowEvents(xdo->xdpy, SyncPointer, CurrentTime);
+  XWindowEvent(xdo->xdpy, screen->root, ButtonReleaseMask, &e);
+  XUngrabPointer(xdo->xdpy, CurrentTime);
+
+  /* If there is no subwindow, then we clicked on the root window */
+  if (e.xbutton.subwindow == 0) {
+    *window_ret = e.xbutton.root;
+  } else {
+     /* Random testing showed that that 'root' always is the same as 'window'
+      * while 'subwindow' is the actual window we clicked on. Confusing... */
+     *window_ret = e.xbutton.subwindow;
+    xdo_window_find_client(xdo, *window_ret, window_ret, XDO_FIND_CHILDREN);
+  }
+  return XDO_SUCCESS;
+}
+
 /* XRaiseWindow is ignored in ion3 and Gnome2. Is it even useful? */
 int xdo_window_raise(const xdo_t *xdo, Window wid) {
   int ret = 0;
@@ -676,8 +780,8 @@ int _xdo_mousebutton(const xdo_t *xdo, Window window, int button, int is_press)
                           xbpe.x_root, xbpe.y_root, &xbpe.x, &xbpe.y, &xbpe.subwindow);
 
     /* Normal behavior of 'mouse up' is that the modifier mask includes
-     * 'ButtonNMotionMask' where N is the button being released. This works the same
-     * way with keys, too. */
+     * 'ButtonNMotionMask' where N is the button being released. This works the
+     * same way with keys, too. */
     if (!is_press) { /* is mouse up */
       switch(button) {
         case 1: xbpe.state |= Button1MotionMask; break;
@@ -702,11 +806,22 @@ int xdo_mousedown(const xdo_t *xdo, Window window, int button) {
   return _xdo_mousebutton(xdo, window, button, True);
 }
 
-int xdo_mouselocation(const xdo_t *xdo, int *x_ret, int *y_ret, int *screen_num_ret) {
+int xdo_mouselocation(const xdo_t *xdo, int *x_ret, int *y_ret,
+                      int *screen_num_ret) {
+  return xdo_mouselocation2(xdo, x_ret, y_ret, screen_num_ret, NULL);
+}
+
+int xdo_mousewindow(const xdo_t *xdo, Window *window_ret) {
+  return xdo_mouselocation2(xdo, NULL, NULL, NULL, window_ret);
+}
+
+int xdo_mouselocation2(const xdo_t *xdo, int *x_ret, int *y_ret,
+                       int *screen_num_ret, Window *window_ret) {
   int ret = False;
   int x = 0, y = 0, screen_num = 0;
   int i = 0;
-  Window dummy_win = 0;
+  Window window = 0;
+  Window root = 0;
   int dummy_int = 0;
   unsigned int dummy_uint = 0;
   int screencount = ScreenCount(xdo->xdpy);
@@ -714,7 +829,7 @@ int xdo_mouselocation(const xdo_t *xdo, int *x_ret, int *y_ret, int *screen_num_
   for (i = 0; i < screencount; i++) {
     Screen *screen = ScreenOfDisplay(xdo->xdpy, i);
     ret = XQueryPointer(xdo->xdpy, RootWindowOfScreen(screen),
-                        &dummy_win, &dummy_win,
+                        &root, &window,
                         &x, &y, &dummy_int, &dummy_int, &dummy_uint);
     if (ret == True) {
       screen_num = i;
@@ -722,10 +837,34 @@ int xdo_mouselocation(const xdo_t *xdo, int *x_ret, int *y_ret, int *screen_num_
     }
   }
 
+  if (window_ret != NULL) {
+    /* Find the client window if we are not root. */
+    if (window != root && window != 0) {
+      int findret;
+      Window client = 0;
+
+      /* Search up the stack for a client window for this window */
+      findret = xdo_window_find_client(xdo, window, &client, XDO_FIND_PARENTS);
+      if (findret == XDO_ERROR) {
+        /* If no client found, search down the stack */
+        findret = xdo_window_find_client(xdo, window, &client, XDO_FIND_CHILDREN);
+      }
+      //fprintf(stderr, "%ld, %ld, %ld, %d\n", window, root, client, findret);
+      if (findret == XDO_SUCCESS) {
+        window = client;
+      }
+    } else {
+      window = root;
+    }
+  }
+  //printf("mouseloc root: %ld\n", root);
+  //printf("mouseloc window: %ld\n", window);
+
   if (ret == True) {
     if (x_ret != NULL) *x_ret = x;
     if (y_ret != NULL) *y_ret = y;
     if (screen_num_ret != NULL) *screen_num_ret = screen_num;
+    if (window_ret != NULL) *window_ret = window;
   }
 
   return _is_success("XQueryPointer", ret == False);
@@ -743,12 +882,26 @@ int xdo_click(const xdo_t *xdo, Window window, int button) {
   return ret;
 }
 
+int xdo_click_multiple(const xdo_t *xdo, Window window, int button,
+                       int repeat, useconds_t delay) {
+  int ret = 0;
+  while (repeat > 0) {
+    ret = xdo_click(xdo, window, button);
+    if (ret != XDO_SUCCESS) {
+      fprintf(stderr, "click failed with %d repeats remaining\n", repeat);
+      return ret;
+    }
+    repeat--;
+    if (repeat > 0) {
+      usleep(delay);
+    }
+  } /* while (repeat > 0) */
+  return ret;
+} /* int xdo_click_multiple */
+
 /* XXX: Return proper code if errors found */
 int xdo_type(const xdo_t *xdo, Window window, char *string, useconds_t delay) {
   int i = 0;
-  //char key = '\0';
-  //XXX: REMOVE //int keycode = 0;
-  //TODO(sissel): REMOVE //int key_index = 0;
 
   /* Since we're doing down/up, the delay should be based on the number
    * of keys pressed (including shift). Since up/down is two calls,
@@ -768,9 +921,24 @@ int xdo_type(const xdo_t *xdo, Window window, char *string, useconds_t delay) {
     key.modmask = 0;
     key.needs_binding = 0;
     if (key.code == 0 && key.symbol == NoSymbol) {
-      fprintf(stderr, "I don't what key produces '%c', skipping.\n",
-              key.key);
-      continue;
+      /* Try the charmap */
+      int kci = 0;
+      //printf("Can't find key %c, checking charmap\n", key.key);
+      for (kci = 0; keysym_charmap[kci].keysym; kci++) {
+        if (key.key == keysym_charmap[kci].key) {
+          key.symbol = XStringToKeysym(keysym_charmap[kci].keysym);
+        }
+      }
+
+      if (key.symbol == NoSymbol) {
+        fprintf(stderr, "I don't what key produces '%c', skipping.\n",
+                key.key);
+        continue;
+      }
+    } else {
+      //printf("Found key for %c\n", key.key);
+      //printf("code: %d\n", key.code);
+      //printf("sym: %s\n", XKeysymToString(key.symbol));
     }
 
     if (key.code > 0) {
@@ -785,7 +953,6 @@ int xdo_type(const xdo_t *xdo, Window window, char *string, useconds_t delay) {
      * around this by binding a new key to the key and using that. */
     if (key.index >= 4) {
       key.needs_binding = 1;
-      //key.index = 0;
     }
 
     if (key.needs_binding == 0) {
@@ -803,7 +970,6 @@ int xdo_type(const xdo_t *xdo, Window window, char *string, useconds_t delay) {
       } 
     }
 
-
     //printf(stderr,
             //"Key '%c' maps to code %d / sym %lu with index %d / mods %d (%s)\n",
             //key.key, key.code, key.symbol, key.index, key.modmask, 
@@ -812,6 +978,7 @@ int xdo_type(const xdo_t *xdo, Window window, char *string, useconds_t delay) {
     //_xdo_send_key(xdo, window, keycode, modstate, True, delay);
     //_xdo_send_key(xdo, window, keycode, modstate, False, delay);
     xdo_keysequence_list_do(xdo, window, &key, 1, True, NULL, delay / 2);
+    key.needs_binding = 0;
     xdo_keysequence_list_do(xdo, window, &key, 1, False, NULL, delay / 2);
 
     /* XXX: Flush here or at the end? or never? */
@@ -882,14 +1049,16 @@ int xdo_keysequence_list_do(const xdo_t *xdo, Window window, charcodemap_t *keys
   for (i = 0; i < nkeys; i++) {
     if (keys[i].needs_binding == 1) {
       KeySym keysym_list[] = { keys[i].symbol };
-      //fprintf(stderr, "Mapping sym %lu to %d\n", keys[i].symbol, scratch_keycode);
+      fprintf(stderr, "Mapping sym %lu to %d\n", keys[i].symbol, scratch_keycode);
       XChangeKeyboardMapping(xdo->xdpy, scratch_keycode, 1, keysym_list, 1);
+      XSync(xdo->xdpy, False);
       /* override the code in our current key to use the scratch_keycode */
       keys[i].code = scratch_keycode;
       keymapchanged = 1;
     }
 
-    //fprintf(stderr, "Sending %c (%d, mods %x)\n", keys[i].key, keys[i].code, *modifier);
+    //fprintf(stderr, "keyseqlist_do: Sending %c %s (%d, mods %x)\n", 
+            //keys[i].key, (pressed ? "down" : "up"), keys[i].code, *modifier);
     _xdo_send_key(xdo, window, &(keys[i]), *modifier, pressed, delay);
 
     if (keys[i].needs_binding == 1) {
@@ -959,18 +1128,20 @@ int xdo_window_get_focus(const xdo_t *xdo, Window *window_ret) {
 int xdo_window_wait_for_focus(const xdo_t *xdo, Window window, int want_focus) {
   Window focuswin = 0;
   int ret;
+  int tries = MAX_TRIES;
   ret = xdo_window_get_focus(xdo, &focuswin);
   if (ret != 0) {
     return ret;
   }
 
-  while (want_focus ? focuswin != window : focuswin == window) {
-    //printf("Current win: %ld, want %ld\n", focuswin, window);
+  while (tries > 0 && 
+         (want_focus ? focuswin != window : focuswin == window)) {
     usleep(30000); /* TODO(sissel): Use exponential backoff up to 1 second */
     ret = xdo_window_get_focus(xdo, &focuswin);
     if (ret != 0) {
       return ret;
     }
+    tries--;
   }
   return 0;
 }
@@ -980,42 +1151,67 @@ int xdo_window_wait_for_focus(const xdo_t *xdo, Window window, int want_focus) {
  * top-level-ish window having focus rather than something you may
  * not expect to be the window having focused. */
 int xdo_window_sane_get_focus(const xdo_t *xdo, Window *window_ret) {
-  int done = 0;
-  Window w;
-  XClassHint classhint;
+  xdo_window_get_focus(xdo, window_ret);
+  xdo_window_find_client(xdo, *window_ret, window_ret, XDO_FIND_PARENTS);
+  return _is_success("xdo_window_sane_get_focus", *window_ret == 0);
+}
 
+int xdo_window_find_client(const xdo_t *xdo, Window window, Window *window_ret,
+                           int direction) {
+  //printf("call findclient: %ld\n", window);
   /* for XQueryTree */
   Window dummy, parent, *children = NULL;
   unsigned int nchildren;
+  Atom atom_wmstate = XInternAtom(xdo->xdpy, "WM_STATE", False);
 
-  xdo_window_get_focus(xdo, &w);
-
+  int done = False;
   while (!done) {
-    Status s;
-    s = XGetClassHint(xdo->xdpy, w, &classhint);
-    //fprintf(stderr, "%d\n", s);
+    if (window == 0) {
+      return XDO_ERROR;
+    }
 
-    if (s == 0) {
-      /* Error. This window doesn't have a class hint */
-      //fprintf(stderr, "no class on: %d\n", w);
-      XQueryTree(xdo->xdpy, w, &dummy, &parent, &children, &nchildren);
+    long items;
+    xdo_getwinprop(xdo, window, atom_wmstate, &items, NULL, NULL);
 
-      /* Don't care about the children, but we still need to free them */
-      if (children != NULL)
-        XFree(children);
-      //fprintf(stderr, "parent: %d\n", parent);
-      w = parent;
+    if (items == 0) {
+      /* This window doesn't have WM_STATE property, keep searching. */
+      XQueryTree(xdo->xdpy, window, &dummy, &parent, &children, &nchildren);
+
+      if (direction == XDO_FIND_PARENTS) {
+        /* Don't care about the children, but we still need to free them */
+        if (children != NULL)
+          XFree(children);
+        window = parent;
+      } else if (direction == XDO_FIND_CHILDREN) {
+        unsigned int i = 0;
+        int ret;
+        done = True; /* recursion should end us */
+        for (i = 0; i < nchildren && !done; i++) {
+          ret = xdo_window_find_client(xdo, children[i], &window, direction);
+          printf("findclient: %ld\n", window);
+          if (ret == XDO_SUCCESS) {
+            *window_ret = window;
+            break;
+          }
+        }
+        if (nchildren == 0) {
+          return XDO_ERROR;
+        }
+        if (children != NULL)
+          XFree(children);
+      } else {
+        fprintf(stderr, "Invalid find_client direction (%d)\n", direction);
+        *window_ret = 0;
+        if (children != NULL)
+          XFree(children);
+        return XDO_ERROR;
+      }
     } else {
-      /* Must XFree the class and name items. */
-      XFree(classhint.res_class);
-      XFree(classhint.res_name);
-
-      done = 1;
+      *window_ret = window;
+      done = True;
     }
   }
-
-  *window_ret = w;
-  return _is_success("xdo_window_sane_get_focus", w == 0);
+  return XDO_SUCCESS;
 }
 
 /* Helper functions */
@@ -1109,6 +1305,8 @@ static void _xdo_populate_charcode_map(xdo_t *xdo) {
       }
 
       xdo->charcodes[idx].key = _keysym_to_char(keybuf);
+      //printf("code: %d[%d] => %c (%s)\n",
+             //i, j, xdo->charcodes[idx].key, XKeysymToString(keysym));
       xdo->charcodes[idx].code = i;
       xdo->charcodes[idx].index = j;
       xdo->charcodes[idx].modmask = _xdo_query_keycode_to_modifier(xdo, i);
@@ -1154,7 +1352,7 @@ char _keysym_to_char(const char *keysym) {
   if (strlen(keysym) == 1)
     return keysym[0];
 
-  return '?';
+  return '\0';
 }
 
 int _xdo_keysequence_to_keycode_list(const xdo_t *xdo, const char *keyseq,
@@ -1316,9 +1514,17 @@ unsigned char *xdo_getwinprop(const xdo_t *xdo, Window window, Atom atom,
   else if (actual_format == 0)
     nbytes = 0;
 
-  *nitems = _nitems;
-  *type = actual_type;
-  *size = actual_format;
+  if (nitems != NULL) {
+    *nitems = _nitems;
+  }
+
+  if (type != NULL) {
+    *type = actual_type;
+  }
+
+  if (size != NULL) {
+    *size = actual_format;
+  }
   return prop;
 }
 
@@ -1375,18 +1581,20 @@ void _xdo_send_key(const xdo_t *xdo, Window window, charcodemap_t *key,
           switch(masks[i]) {
             case ShiftMask:  maskname = "shift"; break;
             case ControlMask: maskname = "control"; break;
+            case Mod1Mask: maskname = "mod1/alt"; break;
             case Mod3Mask: maskname = "mod3"; break;
+            case Mod4Mask: maskname = "mod4/super"; break;
           }
           modkey = _xdo_cached_modifier_to_keycode(xdo, masks[i]);
-          //fprintf(stderr, "Sending modkey for mod %d, %s, (keycode %d) %s\n",
+          //printf("XTEST: Sending modifier key for mod %d, %s, (keycode %d) %s\n",
                  //masks[i], maskname, modkey, is_press ? "down" : "up");
           XTestFakeKeyEvent(xdo->xdpy, modkey, is_press, CurrentTime);
           XSync(xdo->xdpy, False);
         } /* if modestate includes this mask */
       } /* loop over possible masks */
-    }
+    } /* if we have a mask set */ 
 
-    //printf("Sending key %d %s\n", key->code, is_press ? "down" : "up");
+    //printf("XTEST: Sending key %d %s\n", key->code, is_press ? "down" : "up");
     XTestFakeKeyEvent(xdo->xdpy, key->code, is_press, CurrentTime);
     XSync(xdo->xdpy, False);
   } else {
@@ -1599,11 +1807,11 @@ int xdo_window_get_pid(const xdo_t *xdo, Window window) {
   unsigned char *data;
   int window_pid = 0;
 
-  if (_NET_WM_PID == (Atom)-1) {
-    _NET_WM_PID = XInternAtom(xdo->xdpy, "_NET_WM_PID", False);
+  if (atom_NET_WM_PID == (Atom)-1) {
+    atom_NET_WM_PID = XInternAtom(xdo->xdpy, "_NET_WM_PID", False);
   }
 
-  data = xdo_getwinprop(xdo, window, _NET_WM_PID, &nitems, &type, &size);
+  data = xdo_getwinprop(xdo, window, atom_NET_WM_PID, &nitems, &type, &size);
 
   if (nitems > 0) {
     /* The data itself is unsigned long, but everyone uses int as pid values */
@@ -1617,11 +1825,14 @@ int xdo_window_get_pid(const xdo_t *xdo, Window window) {
 int xdo_mouse_wait_for_move_from(const xdo_t *xdo, int origin_x, int origin_y) {
   int x, y;
   int ret = 0;
+  int tries = MAX_TRIES;
 
   ret = xdo_mouselocation(xdo, &x, &y, NULL);
-  while (x == origin_x && y == origin_y) {
+  while (tries > 0 && 
+         (x == origin_x && y == origin_y)) {
     usleep(30000);
     ret = xdo_mouselocation(xdo, &x, &y, NULL);
+    tries--;
   }
 
   return ret;
@@ -1630,12 +1841,132 @@ int xdo_mouse_wait_for_move_from(const xdo_t *xdo, int origin_x, int origin_y) {
 int xdo_mouse_wait_for_move_to(const xdo_t *xdo, int dest_x, int dest_y) {
   int x, y;
   int ret = 0;
+  int tries = MAX_TRIES;
 
   ret = xdo_mouselocation(xdo, &x, &y, NULL);
-  while (x != dest_x && y != dest_y) {
+  while (tries > 0 && (x != dest_x && y != dest_y)) {
     usleep(30000);
     ret = xdo_mouselocation(xdo, &x, &y, NULL);
+    tries--;
   }
 
   return ret;
 }
+
+int xdo_get_desktop_viewport(const xdo_t *xdo, int *x_ret, int *y_ret) {
+  if (_xdo_ewmh_is_supported(xdo, "_NET_DESKTOP_VIEWPORT") == False) {
+    fprintf(stderr,
+            "Your windowmanager claims not to support _NET_DESKTOP_VIEWPORT, "
+            "so I cannot tell you the viewport position.\n");
+    return XDO_ERROR;
+  }
+
+  Atom type;
+  int size;
+  long nitems;
+  unsigned char *data;
+  Atom request = XInternAtom(xdo->xdpy, "_NET_DESKTOP_VIEWPORT", False);
+  Window root = RootWindow(xdo->xdpy, 0);
+  data = xdo_getwinprop(xdo, root, request, &nitems, &type, &size);
+
+  if (type != XA_CARDINAL) {
+    fprintf(stderr, 
+            "Got unexpected type returned from _NET_DESKTOP_VIEWPORT."
+            " Expected CARDINAL, got %s\n",
+            XGetAtomName(xdo->xdpy, type));
+    return XDO_ERROR;
+  }
+
+  if (nitems != 2) {
+    fprintf(stderr, "Expected 2 items for _NET_DESKTOP_VIEWPORT, got %ld\n",
+            nitems);
+    return XDO_ERROR;
+  }
+
+  int *viewport_data = (int *)data;
+  *x_ret = viewport_data[0];
+  *y_ret = viewport_data[1];
+
+  return XDO_SUCCESS;
+}
+
+int xdo_set_desktop_viewport(const xdo_t *xdo, int x, int y) {
+  XEvent xev;
+  int ret;
+  Window root = RootWindow(xdo->xdpy, 0);
+
+  memset(&xev, 0, sizeof(xev));
+  xev.type = ClientMessage;
+  xev.xclient.display = xdo->xdpy;
+  xev.xclient.window = root;
+  xev.xclient.message_type = XInternAtom(xdo->xdpy, "_NET_DESKTOP_VIEWPORT",
+                                         False);
+  xev.xclient.format = 32;
+  xev.xclient.data.l[0] = x;
+  xev.xclient.data.l[1] = y;
+
+  ret = XSendEvent(xdo->xdpy, root, False,
+                   SubstructureNotifyMask | SubstructureRedirectMask, &xev);
+
+  /* XXX: XSendEvent returns 0 on conversion failure, nonzero otherwise.
+   * Manpage says it will only generate BadWindow or BadValue errors */
+  return _is_success("XSendEvent[EWMH:_NET_DESKTOP_VIEWPORT]", ret == 0);
+}
+
+int xdo_window_kill(const xdo_t *xdo, Window window) {
+  int ret;
+  ret = XKillClient(xdo->xdpy, window);
+  return _is_success("XKillClient", ret == 0);
+}
+
+int xdo_get_window_name(const xdo_t *xdo, Window window, 
+                        unsigned char **name_ret, int *name_len_ret,
+                        int *name_type) {
+  if (atom_NET_WM_NAME == (Atom)-1) {
+    atom_NET_WM_NAME = XInternAtom(xdo->xdpy, "_NET_WM_NAME", False);
+  } 
+  if (atom_WM_NAME == (Atom)-1) {
+    atom_WM_NAME = XInternAtom(xdo->xdpy, "WM_NAME", False);
+  }
+  if (atom_STRING == (Atom)-1) {
+    atom_STRING = XInternAtom(xdo->xdpy, "STRING", False);
+  }
+  if (atom_UTF8_STRING == (Atom)-1) {
+    atom_UTF8_STRING = XInternAtom(xdo->xdpy, "UTF8_STRING", False);
+  }
+
+  Atom type;
+  int size;
+  long nitems;
+
+  /**
+   * http://standards.freedesktop.org/wm-spec/1.3/ar01s05.html
+   * Prefer _NET_WM_NAME if available, otherwise use WM_NAME
+   * If no WM_NAME, set name_ret to NULL and set len to 0
+   */
+
+  *name_ret = xdo_getwinprop(xdo, window, atom_NET_WM_NAME, &nitems,
+                             &type, &size);
+  if (nitems == 0) {
+    *name_ret = xdo_getwinprop(xdo, window, atom_WM_NAME, &nitems,
+                               &type, &size);
+  }
+  *name_len_ret = nitems;
+  *name_type = type;
+
+  return 0;
+}
+
+int xdo_window_minimize(const xdo_t *xdo, Window window) {
+  int ret;
+  int screen;
+
+  /* Get screen number */
+  XWindowAttributes attr;
+  XGetWindowAttributes(xdo->xdpy, window, &attr);
+  screen = XScreenNumberOfScreen(attr.screen);
+
+  /* Minimize it */
+  ret = XIconifyWindow(xdo->xdpy, window, screen);
+  return _is_success("XIconifyWindow", ret == 0);
+}
